#include "version.h"

#ifdef USE_SHADING
#include "wl_def.h"
#include "wl_shade.h"

typedef struct {
    uint8_t destRed, destGreen, destBlue;   // values between 0 and 255
    uint8_t fogStrength;
} shadedef_t;

shadedef_t shadeDefs[] = {
    {   0,   0,   0, LSHADE_NOSHADING },
    {   0,   0,   0, LSHADE_NORMAL },
    {   0,   0,   0, LSHADE_FOG },
    {   0,  40,   0, LSHADE_FOG },
    { 160, 160, 160, LSHADE_FOG },
    { 80,   0,   0, 7 },
    {  0,   0,   0, 15},
};

uint8_t shadetable[SHADE_COUNT][256];
uint8_t indoortable[SHADE_COUNT][256];
int SHADE_flag, ISHADE_flag;
#ifdef USE_FEATUREFLAGS

// The lower 8-bit of the upper left tile of every map determine
// the used shading definition of shadeDefs.
static inline int GetShadeDefID()
{
    int shadeID = ffDataTopLeft & 0x00ff;
    assert(shadeID >= 0 && shadeID < lengthof(shadeDefs));
    return shadeID;
}

static inline int GetIndoorShadeDefID()
{
    int shadeID = ffDataTopLeft >> 8;
    assert(shadeID >= 0 && shadeID < lengthof(shadeDefs));
    return shadeID;
}

#else

static int GetShadeDefID()
{
    int shadeID;
    switch(gamestate.mapon)
    {
        case  12: shadeID = 3; break;
        case  33: shadeID = 4; break;
        case  35: shadeID = 4; break;
        case  36: shadeID = 4; break;
        case  37: shadeID = 4; break;
        case  43: shadeID = 5; break;
        case  44: shadeID = 5; break;
        case  45: shadeID = 5; break;
        default: shadeID = 2; break;
    }
    assert(shadeID >= 0 && shadeID < lengthof(shadeDefs));
    return shadeID;
}

#endif

// Returns the palette index of the nearest matching color of the
// given RGB color in given palette
byte GetColor(byte red, byte green, byte blue, SDL_Color *palette)
{
    byte mincol = 0;
    double mindist = 200000.F, curdist, DRed, DGreen, DBlue;

    SDL_Color *palPtr = palette;

    for(int col = 0; col < 256; col++, palPtr++)
    {
        DRed   = (double) (red   - palPtr->r);
        DGreen = (double) (green - palPtr->g);
        DBlue  = (double) (blue  - palPtr->b);
        curdist = DRed * DRed + DGreen * DGreen + DBlue * DBlue;
        if(curdist < mindist)
        {
            mindist = curdist;
            mincol = (byte) col;
        }
    }
    return mincol;
}

#ifdef TRANSPARENTRAIN
// Merges two colours based on their strength and returns the closest palette colour
// If only we had 32-bit colour... *scratches chin*
byte MergeColour(byte col1, byte col2, float c1s, float c2s)
{
    SDL_Color *pal = gamepal;
    byte r, g, b;

    r = (byte) ((((pal + col1)->r)*c1s + ((pal + col2)->r)*c2s) / (c1s + c2s));
    g = (byte) ((((pal + col1)->g)*c1s + ((pal + col2)->g)*c2s) / (c1s + c2s));
    b = (byte) ((((pal + col1)->b)*c1s + ((pal + col2)->b)*c2s) / (c1s + c2s));

    return GetColor(r, g, b, gamepal);
}
#endif
// Fade all colors in 32 steps down to the destination-RGB
// (use gray for fogging, black for standard shading)
void GenerateIndoorShadeTable(byte destRed, byte destGreen, byte destBlue,
                        SDL_Color *palette, int fog)
{
    double curRed, curGreen, curBlue, redStep, greenStep, blueStep;
    SDL_Color *palPtr = palette;

    // Set the fog-flag
    ISHADE_flag=fog;

    // Color loop
    for(int i = 0; i < 256; i++, palPtr++)
    {
        // Get original palette color
        curRed   = palPtr->r;
        curGreen = palPtr->g;
        curBlue  = palPtr->b;

        // Calculate increment per step
        redStep   = ((double) destRed   - curRed)   / (SHADE_COUNT + 8);
        greenStep = ((double) destGreen - curGreen) / (SHADE_COUNT + 8);
        blueStep  = ((double) destBlue  - curBlue)  / (SHADE_COUNT + 8);

        // Calc color for each shade of the current color
        for (int shade = 0; shade < SHADE_COUNT; shade++)
        {
            indoortable[shade][i] = GetColor((byte) curRed, (byte) curGreen, (byte) curBlue, palette);

            // Inc to next shade
            curRed   += redStep;
            curGreen += greenStep;
            curBlue  += blueStep;
        }
    }
}

// Fade all colors in 32 steps down to the destination-RGB
// (use gray for fogging, black for standard shading)
void GenerateShadeTable(byte destRed, byte destGreen, byte destBlue,
                        SDL_Color *palette, int fog)
{
    double curRed, curGreen, curBlue, redStep, greenStep, blueStep;
    SDL_Color *palPtr = palette;

    // Set the fog-flag
    SHADE_flag=fog;

    // Color loop
    for(int i = 0; i < 256; i++, palPtr++)
    {
        // Get original palette color
        curRed   = palPtr->r;
        curGreen = palPtr->g;
        curBlue  = palPtr->b;

        // Calculate increment per step
        redStep   = ((double) destRed   - curRed)   / (SHADE_COUNT + 8);
        greenStep = ((double) destGreen - curGreen) / (SHADE_COUNT + 8);
        blueStep  = ((double) destBlue  - curBlue)  / (SHADE_COUNT + 8);

        // Calc color for each shade of the current color
        for (int shade = 0; shade < SHADE_COUNT; shade++)
        {
            shadetable[shade][i] = GetColor((byte) curRed, (byte) curGreen, (byte) curBlue, palette);

            // Inc to next shade
            curRed   += redStep;
            curGreen += greenStep;
            curBlue  += blueStep;
        }
    }
}

void NoShading()
{
    for(int shade = 0; shade < SHADE_COUNT; shade++)
        for(int i = 0; i < 256; i++)
            shadetable[shade][i] = i;
}

#ifdef ALULIGHTING
void NoIndoorShading()
{
    for(int shade = 0; shade < SHADE_COUNT; shade++)
        for(int i = 0; i < 256; i++)
            indoortable[shade][i] = i;
}
#endif

void InitLevelShadeTable()
{
    shadedef_t *shadeDef = &shadeDefs[GetShadeDefID()];
#ifdef ALULIGHTING
    if (GetBRFlags() & BR_UPPERINDOORSHADE && !(GetBRFlags() & BR_NOSHADEINDOOR))
    {
        shadedef_t *indoorDef = &shadeDefs[GetIndoorShadeDefID()];
        GenerateIndoorShadeTable(indoorDef->destRed, indoorDef->destGreen, indoorDef->destBlue, gamepal, indoorDef->fogStrength);
    }
    else if (GetBRFlags() & BR_NOSHADEINDOOR)
        NoIndoorShading();
    else
        GenerateIndoorShadeTable(shadeDef->destRed, shadeDef->destGreen, shadeDef->destBlue, gamepal, shadeDef->fogStrength);
#endif

    if(shadeDef->fogStrength == LSHADE_NOSHADING)
        NoShading();
    else
        GenerateShadeTable(shadeDef->destRed, shadeDef->destGreen, shadeDef->destBlue, gamepal, shadeDef->fogStrength);
}

#ifdef ALULIGHTING
int GetShade(int scale, bool inside)
#else
int GetShade(int scale)
#endif
{
	int flag = SHADE_flag;
#ifdef ALULIGHTING
	if (inside)
		flag = ISHADE_flag;
#endif
    int shade = (scale >> 1) / (((viewwidth * 3) >> 8) + 1 + flag);  // TODO: reconsider this...
    if(shade > 32) shade = 32;
    else if(shade < 1) shade = 1;
    shade = 32 - shade;

    return shade;
}

#endif
